/*
 * Container.tcc    Step 6, Iterators.
 */
#ifndef CONTAINER_TCC
#define CONTAINER_TCC

#include "Container.h"
#include <algorithm>
#include <cassert>
#include <stdexcept>
#include <utility>
using namespace std;

/*
 * compute_capacity_() [STEP 4]
 * Computes the next capacity according to allocation sceme.
 * Note: size_type(-1) is instead for std::numeric_limits<size_type>::max().
 */
template <typename T>
typename Container<T>::size_type
Container<T>::
compute_capacity_() const
{
   if (capacity() == size_type(-1))
      throw std::length_error{ "Container capacity is exhausted" };
   else if (capacity() >= size_type(-1) / 2)
      return size_type(-1);
   return capacity() == 0 ? 1 : 2 * capacity();
}

/*
 * allocate_(n) [STEP 1]
 * Allocates memory for n objects of type T using array new. All elements will
 * be value-initialized (defauilt initialized if T i a class type).
 */
template <typename T>
typename Container<T>::pointer 
Container<T>::
allocate_(size_type n)
{
   if (n > max_size())
      throw std::bad_alloc();

   if (n > 0)
      return new T[n];
   else
      return nullptr;
}

/*
 * deallocate_(pointer) [STEP 1]
 * Deallocates memory using array delete. All elements will be destroyed.
 */
template <typename T>
void
Container<T>::
deallocate_(pointer start)
{
   delete[] start;  // can't fail
}

/*
 * allocate_and_copy_(const Container&) [STEP 2]
 * Copies the content of a container to new storage (shrink-to-fit). If a failure
 * is encountered during copying elements from current to the new storage, the new
 * storage is deallocated and all initialized elements destroyed when doing that.
 */
template <typename T>
typename Container<T>::pointer
Container<T>::
allocate_and_copy_(const Container& c)
{
   if (c.size() == 0)
      return nullptr;

   pointer new_start = allocate_(c.size());

   try
   {
      std::copy(c.begin(), c.end(), new_start);
   }
   catch (...)
   {
      deallocate_(new_start);  // can't fail
      throw;
   }

   return new_start;
}

/*
 * resize(const_reference) [STEP 4]
 * Resizes storage according to capacity scheme, insert x, and copy excisting 
 * elements to the new storage.
 */
template <typename T>
void
Container<T>::
resize_(const_reference x)
{
   size_type old_size = size();
   size_type new_capacity = compute_capacity_();
   pointer   new_start = allocate_(new_capacity);

   try
   {
      *(new_start + old_size) = x;
      std::copy(begin(), end(), new_start);
   }
   catch (...)
   {
      deallocate_(new_start);  // can't fail
      throw;
   }

   delete[] start_;  // can't fail
   start_ = new_start;
   finish_ = start_ + old_size;
   end_of_storage_ = start_ + new_capacity;

}

/*
 * Constructor initializing to size n [STEP 1]
 * If n < 1 the container is initialzed to empty (n cannot be negative), else storage
 * is allocated for n elements, all objects will be value initialized.
 */
template <typename T>
Container<T>::
Container(const size_type n)
{
   if (n > 0)
   {
      start_ = allocate_(n);
      finish_ = start_ + n;
      end_of_storage_ = finish_;
   }
}

/*
 * Copy constructor [STEP 2]
 */
template <typename T>
Container<T>::
Container(const Container& other)
{
   if (other.size() > 0)
   {
      start_ = allocate_and_copy_(other);
      finish_ = start_ + other.size();
      end_of_storage_ = finish_;
   }
}

/*
 * Move constructor [STEP 2]
 * Destination container is initialized to empty, then content is swapped,
 * source container becomes empty.
 */
template <typename T>
Container<T>::
Container(Container&& other) noexcept
{
   swap(other);  // can't fail
}

/*
 * Destructor [STEP 1]
 * Deallocates storage, all elements are destroyed before that.
 */
template <typename T>
Container<T>::
~Container()
{
   deallocate_(start_);  // can't fail
}

/*
 * Copy assignment operator [STEP 2, modified STEP 3]
 * Makes a copy of other (shrink-to-fit), as a temporary, and swaps content
 * with *this.
 */
template <typename T>
Container<T>&
Container<T>::
operator=(const Container& other) &
{
   if (this != &other)
      Container{ other }.swap(*this);
   return *this;  // safe, no copy involved
}

/*
 * Move assignment operator [introduced STEP 2, modified STEP 3]
 * Deallocates storage and swaps content with other, which is then made empty.
 */
template <typename T>
Container<T>&
Container<T>::
operator=(Container&& other) & noexcept
{
   if (this != &other)
   {
      clear();
      swap(other);
   }
   return *this;   // safe, no copy involved
}

/*
 * size() [STEP 1, modified STEP 6]
 */
template <typename T>
typename Container<T>::size_type
Container<T>::
size() const noexcept
{
   return size_type(end() - begin());
}

/*
 * max_size() [STEP 1]
 * Largest chunk of memory, in bytes, is equal to the largest value for size_type.
 * Note: size_type(-1) is used instead of std::numeric_limits<size_type>::max().
 */
template <typename T>
typename Container<T>::size_type
Container<T>::
max_size() const noexcept
{
   return size_type(-1) / sizeof(T);
}

/*
 * capacity() [STEP 1]
 */
template <typename T>
typename Container<T>::size_type
Container<T>::
capacity() const noexcept
{
   return size_type(end_of_storage_ - start_);
}

/*
 * empty() [STEP 1, modified STEP 6]
 */
template <typename T>
bool 
Container<T>::
empty() const noexcept
{
   return begin() == end();
}

/*
 * clear() [STEP 3]
 * Destroys all elements, empty() shall return true.
 */
template <typename T>
void 
Container<T>::
clear() noexcept
{
   deallocate_(start_);  // can't fail
   start_ = nullptr;
   finish_ = nullptr;
   end_of_storage_ = nullptr;
}

/*
 * swap(Container&) [STEP 3]
 */
template <typename T>
void 
Container<T>::
swap(Container& other) noexcept
{
   // Swapping pointers can't fail
   std::swap(start_, other.start_);
   std::swap(finish_, other.finish_);
   std::swap(end_of_storage_, other.end_of_storage_);
}

/*
 * swap(Container&, Container&) [STEP 3]
 */
template <typename T>
void swap(Container<T>& a, Container<T>& b) noexcept
{
   a.swap(b);  // can't fail
}

/*
 * push_back(const_reference) [STEP 4]
 */
template <typename T>
void 
Container<T>::
push_back(const_reference x)
{
   if (size() == capacity())
      resize_(x);
   else
      *finish_ = x;
   ++finish_;
}

/*
 * back() [STEP 4, modified STEP 6]
 * For non-const Container, should not throw according to standard. Implemented
 * according to operational semantics stated in reference manual.
 */
template <typename T>
typename Container<T>::reference 
Container<T>::
back()
{
   auto tmp = end();
   --tmp;
   return *tmp;
}

/*
 * back() [STEP 4, modified STEP 6]
 * For const Container, should not throw according to standard. Implemented
 * according to operational semantics stated in reference manual.
 */
template <typename T>
typename Container<T>::const_reference
Container<T>::
back() const
{
   auto tmp = end();
   --tmp;
   return *tmp;
}

/*
 * pop_back() [STEP 4]
 * Destroys the last element. Requires empty() shall be false.
 */
template <typename T>
void
Container<T>::
pop_back() noexcept
{
   --finish_;
   finish_->~T();
}

/*
 * Iterator functions [STEP 6]
 */
template <typename T>
typename Container<T>::iterator
Container<T>::
begin() noexcept
{
   return iterator(start_);
}

template <typename T>
typename Container<T>::const_iterator
Container<T>::
begin() const noexcept
{
   return const_iterator(start_);
}

template <typename T>
typename Container<T>::iterator
Container<T>::
end() noexcept
{
   return iterator(finish_);
}

template <typename T>
typename Container<T>::const_iterator
Container<T>::
end() const noexcept
{
   return const_iterator(finish_);
}

template <typename T>
typename Container<T>::reverse_iterator
Container<T>::
rbegin() noexcept
{
   return reverse_iterator(end());
}

template <typename T>
typename Container<T>::const_reverse_iterator
Container<T>::
rbegin() const noexcept
{
   return const_reverse_iterator(end());
}

template <typename T>
typename Container<T>::reverse_iterator
Container<T>::
rend() noexcept
{
   return reverse_iterator(begin());
}

template <typename T>
typename Container<T>::const_reverse_iterator
Container<T>::
rend() const noexcept
{
   return const_reverse_iterator(begin());
}

template <typename T>
typename Container<T>::const_iterator
Container<T>::
cbegin() const noexcept
{
   return const_iterator(start_);
}

template <typename T>
typename Container<T>::const_iterator
Container<T>::
cend() const noexcept
{
   return const_iterator(finish_);
}

template <typename T>
typename Container<T>::const_reverse_iterator
Container<T>::
crbegin() const noexcept
{
   return const_reverse_iterator(end());
}

template <typename T>
typename Container<T>::const_reverse_iterator
Container<T>::
crend() const noexcept
{
   return const_reverse_iterator(begin());
}

#endif
